// Created by: Daniel Sanchez
// Modified by: Tomás Pérez

#include <fstream>
#include <dirent.h> 
#include <opencv/cv.h>
#include <opencv/highgui.h>
#include <sstream>
#include "detectcascadas.h"
#include "imgrotation.h"
#include "dorientation.h"
#include "svm.h"
#include "main.h"

// Override cmp functions for CvPoint and set of CvPoints

inline bool operator<(const CvPoint& lhs, const CvPoint& rhs) {
    return (lhs.x < rhs.x) || (lhs.y < rhs.y);
}

inline bool operator<(const set<CvPoint>& lhs, const set<CvPoint>& rhs) {
    return (lhs.size() < rhs.size());
}

// Array with every trained model
svm_model * MODEL[NUM_CAS];

float mmean = 0.0;
float final = 0.0;
int iiindex = 0;

// Shows type of error

void showError(const char *s) {
    perror(s);
    exit(EXIT_FAILURE);
}

void cargaModelos() {
    // load array with the models
    for (int x = 0; x < NUM_CAS; x++) {
        MODEL[x] = svm_load_model(modelFiles[x].c_str());
        cout << "Loading model " << x << endl;
    }
}

// Converts an image into a black image

void computeBlackImage(IplImage *image) {
    CvScalar value; // Accesses to matrix elements
    // Accesses to matrix elements and sets values to RGB:0 -> black
    for (int i = 0; i < image->height; i++) {
        for (int j = 0; j < image->width; j++) {
            value = cvGet2D(image, i, j);
            value.val[0] = 0;
            cvSet2D(image, i, j, value);
        }
    }
}

// Sliding window process

void runWindowing(IplImage *image, String imageName) {
    int size = SIZE; // Initial region size, same as cascades region size
    int posY = 0; // Position Y of region
    int posX = 0; // Position X of region
    int height = image->height; // Image height
    int width = image->width; // Image width

    int vect[NUM_CAS]; // Returns result ECOC encoding to this vector

    ostringstream name(""); // Composes name of image

    IplImage *region = NULL; // Region to open

    IplImage * articulations[N_ROWS];

    CvScalar value; // Gets values from any image
    RegionData regionData; // Data for every region
    vector<RegionData> regions; // Contains all regions data from image

    // Does the same but for each limb
    // Instances images in channel 3 just to have "pretty" images and use them for the report
    for (int i = 0; i < N_ROWS; i++) {
        articulations[i] = cvCreateImage(cvGetSize(image), IPL_DEPTH_16U, 1);
        computeBlackImage(articulations[i]);
    }

    // Extracts information from every region
    for (size; size < LIMIT; size = size * GEOMETRIC_FACTOR) {
        cout << "Size: " << size << endl;
        for (posY = 0; (posY + size) < height; posY += STEP) {
            cout << "Y axis " << posY << " / " << (height - size - 1) << " (" << size << ") " << endl;
            for (posX = 0; (posX + size) < width; posX += STEP) {
                // Rotates and creates region
                region = rotateAndCut(image, posX, posY, size);
                // Resize image to get HOG descriptor
                region = img_resize(region, 32, 32);
                // Sends region to the cascades and gets the possibles
                // assignments for every limb
                obtenerVector(region, vect, MODEL);
                // Frees memory
                cvReleaseImage(&region);
                // Contrasts result with ECOC matrix
                computeEcocData(vect, &regionData.data);

                // Fills the other information
                regionData.posX = posX;
                regionData.posY = posY;
                regionData.size = size;

                // Adds to the list
                regions.push_back(regionData);

                // Votes the region of the image regard to the current region
                // Votes for each limb
                for (int k = 0; k < N_ROWS; k++) {
                    if (regionData.data.name.compare(TYPE[k]) == 0) {
                        for (int i = posY; i < (posY + size); i++) {
                            for (int j = posX; j < (posX + size); j++) {
                                value = cvGet2D(articulations[k], i, j);
                                value.val[0]++;
                                cvSet2D(articulations[k], i, j, value);
                            }
                        }
                    }
                }
            }
        }
    }

    // Save an image by articulation
    for (int k = 0; k < N_ROWS; k++) {
        name << MAP_PATH << extractName(imageName) << "_" << MASK_ID[k] << MP_SUFFIX;
        cvSaveImage(name.str().c_str(), articulations[k]);
        name.str("");
    }

    // Frees memory
    for (int k = 0; k < N_COLUMNS; k++) cvReleaseImage(&articulations[k]);
    regions.clear();
    name.str("");
}

// Fills struct 'data' with the articulation name 

void getArticulation(struct EcocData *data) {
    // Name of row type
    ostringstream ecocRow("");

    // Converts int[] to String
    for (int i = 0; i < N_COLUMNS; i++) ecocRow << data->row[i];
    if (ecocRow.str().compare("11000") == 0) {
        data->name = TYPE[0];
    } else if (ecocRow.str().compare("1-1000") == 0) {
        data->name = TYPE[1];
    } else if (ecocRow.str().compare("-10110") == 0) {
        data->name = TYPE[2];
    } else if (ecocRow.str().compare("-101-10") == 0) {
        data->name = TYPE[3];
    } else if (ecocRow.str().compare("-10-101") == 0) {
        data->name = TYPE[4];
    } else if (ecocRow.str().compare("-10-10-1") == 0) {
        data->name = TYPE[5];
    }
}

// Fills struct 'data' with a suitable ECOC row(regard to the type of articulation) and a vector of distances

void computeEcocData(int currentRow[], struct EcocData * data) {
    int correctIndex; // Index of the suitable ECOC row
    int mapIndex = 0; // Index used in map structure
    float distance; // Current distance
    float min = LONG_MAX; // Minimum distance
    float sumOnes = 0.0; // Sums +1 and -1 from ECOC row
    float sumComponents = 0.0; // Sums results from a row components plus ECOC row components

    float distances[N_ROWS]; // Distances regard to ECOC

    map<int, int> equals; // Map contains same distances

    // Loss-Weighted process
    for (int i = 0; i < N_ROWS; i++) {
        // Computes members of division
        for (int j = 0; j < N_COLUMNS; j++) {
            sumComponents -= currentRow[j] * ECOC[i][j] * WEIGHTS[j];
            sumOnes += abs(ECOC[i][j]);
        }

        // Computes distance and saves it into array
        distance = sumComponents / sumOnes;
        distances[i] = distance;

        // Takes the lowest distances and ECOC row index
        if (distance < min) {
            min = distance;
            correctIndex = i;
        }
        // Reset
        sumComponents = 0;
        sumOnes = 0;
    }

    // Exists the case that there are distances which are equals and the lowest 
    // distance can be one of those
    // By default, the first is always chosen, to solve it we choose a random one
    for (int i = 0; i < N_ROWS; i++) {
        if (min == distances[i]) {
            equals[mapIndex] = i;
            mapIndex++;
        }
    }

    // Takes a random value if the lowest value is repeated many times
    if (equals.size() > 1) correctIndex = equals[rand() % equals.size()];

    // Gets distances and correct ECOC row
    for (int i = 0; i < N_ROWS; i++) data->confidences[i] = distances[i];
    for (int i = 0; i < N_COLUMNS; i++) data->row[i] = ECOC[correctIndex][i];
    // Gets type of articulation
    getArticulation(data);
    // Cleans map
    equals.clear();
}

// Sets pixel color regard to the type of articulation

void setPixel(int type, int posY, int posX, IplImage * image) {
    CvScalar value = cvGet2D(image, posY, posX);
    //BGR
    switch (type) {
            // torso
        case 0: // Green
            value.val[1]++;
            break;
        case 1: // Red
            // Cabeza
            value.val[2]++;
            break;
            // Antebrazos
        case 2: // Blue
            value.val[0]++;
            break;
            // Brazos
        case 3: // Yellow
            value.val[1]++;
            value.val[2]++;
            break;
            // Muslos 
        case 4: // Violet
            value.val[0]++;
            value.val[2]++;
            break;
            // Piernas 
        case 5: // Cyan
            value.val[0]++;
            value.val[1]++;
            break;
        default:
            cout << "Bad assignment in pixel of image of probability" << endl;
    }

    cvSet2D(image, posY, posX, value);
}

// Extracts the name of a file from a path
// Also, image name may have white spaces which must be deleted

String extractName(String path) {
    String name;

    name = path.substr(0, path.rfind("."));
    name = name.substr(name.rfind("\\") + 1, name.length());
    replace(name.begin(), name.end(), ' ', '_');

    return name;
}