/*
 * dorientation.cpp
 *
 *  Created on: 21/04/2012
 *      Author: juancarlos
 *      Mail  : juancarloslinux@gmail.com
 */
#include <iostream>
#include <cmath>
#include <opencv/cv.h>
#include <opencv/highgui.h>
#include "dorientation.h"

using namespace std;
using namespace cv;

/**
 * Constantes
 */
enum ConvolutionType {
	/* Return the full convolution, including border */
	CONVOLUTION_FULL,

	/* Return only the part that corresponds to the original image */
	CONVOLUTION_SAME,

	/* Return only the submatrix containing elements that were not influenced by the border */
	CONVOLUTION_VALID
};

/**
 * Realiza la convolución de la imagen con el kernel
 */
static void conv2(const Mat &img, const Mat& kernel, ConvolutionType type,
		Mat& dest) {
	Mat source = img;
	if (CONVOLUTION_FULL == type) {
		source = Mat();
		const int additionalRows = kernel.rows - 1, additionalCols = kernel.cols
				- 1;
		copyMakeBorder(img, source, (additionalRows + 1) / 2,
				additionalRows / 2, (additionalCols + 1) / 2,
				additionalCols / 2, BORDER_CONSTANT, Scalar(0));
	}

	Point anchor(kernel.cols - kernel.cols / 2 - 1,
			kernel.rows - kernel.rows / 2 - 1);
	int borderMode = BORDER_CONSTANT;

	Mat flipped;
	flip(kernel, flipped, -1);

	filter2D(source, dest, CV_32FC1, flipped, anchor, 0, borderMode);

	if (CONVOLUTION_VALID == type) {
		dest =
				dest.colRange((kernel.cols - 1) / 2,
						dest.cols - kernel.cols / 2).rowRange(
						(kernel.rows - 1) / 2, dest.rows - kernel.rows / 2);
	}
}

/**
 * Calcula la secuencia de elementos en función del valor sigma
 */
static vector<int> calcularSecuencia(int sigma) {
	int min_rang = floor(-3.0 * sigma + 0.5);
	int max_rang = floor(3.0 * sigma + 0.5);

	vector<int> x;
	int i = min_rang;
	while (i <= max_rang) {
		x.push_back(i);
		i++;
	}
	return x;
}

/**
 * Calcula la mascara gausiana
 */
static vector<double> calcularGausiana(vector<int> x, int sigma) {
	double sum = 0;
	vector<double> G(x.size());
	for (unsigned int i = 0; i < G.size(); i++) {
		G[i] = exp(-(pow(double(x[i]), 2) / (2 * pow(double(sigma), 2))));
		sum = sum + G[i];
	}

	for (unsigned int i = 0; i < G.size(); i++)
		G[i] = G[i] / sum;

	return G;
}

/**
 * Calcula la derivada
 */
static vector<double> calcularDerivada(vector<int> x, double sigma) {
	vector<double> D(x.size());
	for (unsigned int i = 0; i < x.size(); i++) {
		D[i] = -2 * (x[i] * exp(-double(x[i] * x[i]) / (2 * (sigma * sigma))))
				/ (sqrt(2 * CV_PI) * pow(sigma, 3));
	}
	return D;
}

/**
 * Calcula la magnitud del gradiente
 *
 * OJO!!! La imagen tiene que ser en Blanco y Negro
 */
static void calcularMagnitudGradiente(IplImage *img, int sigma, CvMat **imgMag,
		CvMat **imgDir) {

	vector<int> x = calcularSecuencia(sigma);
	vector<double> G = calcularGausiana(x, sigma);
	vector<double> D = calcularDerivada(x, sigma);

	Mat img1(img->height, img->width, CV_32FC1), imgDx(img->height, img->width,
			CV_32FC1);
	Mat img2(img->height, img->width, CV_32FC1), imgDy(img->height, img->width,
			CV_32FC1);

	// Convolucionamos
	conv2(Mat(img, CV_32FC1), Mat(D).t(), CONVOLUTION_SAME, img1);
	conv2(img1, Mat(G, CV_32FC1), CONVOLUTION_SAME, imgDx);
	conv2(Mat(img, CV_32FC1), Mat(D, CV_32FC1), CONVOLUTION_SAME, img2);
	conv2(img2, Mat(G).t(), CONVOLUTION_SAME, imgDy);

	// Transformamos a matriz
	CvMat imgDxCnv = imgDx;
	CvMat imgDyCnv = imgDy;
	*imgMag = cvCreateMat(imgDxCnv.rows, imgDxCnv.cols, CV_32FC1);
	*imgDir = cvCreateMat(imgDxCnv.rows, imgDxCnv.cols, CV_32FC1);

	// Calculamos lo valores de la magnitud y la dirección
	double valDx;
	double valDy;
	for (unsigned int i = 0; i < (unsigned) imgDxCnv.rows; i++) {
		for (unsigned int j = 0; j < (unsigned) imgDxCnv.cols; j++) {
			valDx = cvGetReal2D(&imgDxCnv, i, j);
			valDy = cvGetReal2D(&imgDyCnv, i, j);
			cvSetReal2D(*imgMag, i, j, sqrt(pow(valDx, 2) + pow(valDy, 2)));
			cvSetReal2D(*imgDir, i, j, atan2(valDy, valDx));
		}
	}
}

/**
 * Calcula la orientacion dominante de una imagen de entrada
 */
double calculateDominantOrientation(IplImage *img) {

	CvMat *imgMag;
	CvMat *imgDir;
	int sigma = 1; // Sigma prefijada
	int nbins = 90;
	vector<double> his(nbins, 0);
	double radio = img->height / 5.0;
	double radio2 = pow(radio, 2);
	int mitad = floor(img->height / 2);

	// Calculamos la magnitud del gradiente
	calcularMagnitudGradiente(img, sigma, &imgMag, &imgDir);
        
	double pix;
	double pixMag;
	double pixOri;
	int index = 0;
	double max = 0.0;
	for (unsigned int i = 1; i <= (unsigned) imgDir->rows; i++) {
        for (unsigned int j = 1; j <= (unsigned) imgDir->cols; j++) {
            pix = cvGetReal2D(imgDir, i - 1, j - 1);
            pixMag = cvGetReal2D(imgMag, i - 1, j - 1);
            pixOri = floor((nbins - 1) * (pix + CV_PI) / (2 * CV_PI)) + 1;

            // Calculamos el cuadrado
            if ((j - mitad) * (j - mitad) + (i - mitad) * (i - mitad) < radio2) {
                // COMPROBACION DE RANGO!!!
                if (int(pixOri) > 0 && int(pixOri) <= nbins) {
                    his[int(pixOri) - 1] = his[int(pixOri) - 1] + pixMag;
                    // Actualizamos si es el maximo
                    if (his[int(pixOri) - 1] > max) {
                        index = int(pixOri);
                        max = his[int(pixOri) - 1];
                    }
                }

            }
        }
    }
    cvReleaseMat(&imgMag);
    cvReleaseMat(&imgDir);
    return (2 * CV_PI * index) / nbins;
}

